#ifndef _H_MCU_COLLECT_TYPED_HASH_MAP
#define _H_MCU_COLLECT_TYPED_HASH_MAP

#include <stddef.h>
#include <stdint.h>
#include <stdlib.h>
#include "mcutil/mcutil.h"

#define MCU_TYPED_HM_DEFINE(name_prefix, key_t, value_t)    \
typedef struct name_prefix##_hm_node_s {                    \
    uint8_t key_hash[16];                                   \
    key_t key;                                              \
    value_t value;                                          \
    struct mcu_hash_map_node_s* next;                       \
} name_prefix##_hm_node_t;                                  \
                                                            \
typedef struct name_prefix##_hm_s {                         \
    name_prefix##_hm_node_t* nodes;                         \
} name_prefix##_hm_t;


#define MCU_TYPED_HM_IMPL(pfx, key_t, value_t)              \
MCU_API pfx##_hm_t mcu_hm_new() {                           \
    return (pfx##_hm_t){                                    \
        .nodes = MCU_ALLOC(sizeof(pfx##_hm_t)),             \
    };                                                      \
}                                                           \
                                                            \
MCU_API void pfx##_hm_insert(                               \
        pfx##_hm_t* hm,                                     \
        const key_t* key,                                   \
        const value_t* value) {                             \
    assert(hm && "HashMap ptr is null");                    \
    mcu_md5_context ctx = {0};                              \
    mcu_md5_init(&ctx);                                     \
    mcu_md5_digest(&ctx, (void*)key, sizeof(key_t));        \
    uint8_t key_res[16];                                    \
    mcu_md5_output(&ctx, key_res);                          \
                                                            \
    if (!(hm)->nodes)                                       \
        (hm)->nodes = MCU_ALLOC(sizeof(pfx##_hm_node_t));   \
                                                            \
    pfx##_hm_node_t* node = hm->nodes;                      \
    while (true) {                                          \
        if (!node->next) {                                  \
            node->next = MCU_ALLOC(sizeof(pfx##_hm_node_t));\
            node->next->next = NULL;                        \
            node->next->key = key;                          \
            memcpy(node->next->key_hash, key_res, 16);      \
            node->next->value = (void*)value;               \
            return;                                         \
        }                                                   \
        node = node->next;                                  \
    }                                                       \
}                                                           \
                                                            \
MCU_API void* pfx##_hm_get(                                 \
        pfx##_hm_t* hm,                                     \
        const key_t key,                                    \
    ) {                                                     \
    assert(hm && "HashMap ptr is null");                    \
    mcu_md5_context ctx = {0};                              \
    mcu_md5_init(&ctx);                                     \
    mcu_md5_digest(&ctx, (void*)key, sizeof(key_t));        \
    uint8_t key_res[16];                                    \
    mcu_md5_output(&ctx, key_res);                          \
                                                            \
    if (!(hm)->nodes)                                       \
        (hm)->nodes = MCU_ALLOC(sizeof(pfx##_hm_node_t));   \
    pfx##_hm_node_t* node = hm->nodes;                      \
    while (true) {                                          \
        if (memcmp(node->key, key_res, 16) == 0) {          \
            return node->value;                             \
        } else {                                            \
            if (!node->next) return NULL;                   \
            node = node->next;                              \
        }                                                   \
    }                                                       \
}                                                           \
                                                            \
MCU_API void* pfx##_hm_remove(                              \
        pfx##_hm_t* hm,                                     \
        const key_t key,                                    \
        const size_t key_len                                \
    ) {                                                     \
    assert(hm && "HashMap ptr is null");                    \
    mcu_md5_context ctx = {0};                              \
    mcu_md5_init(&ctx);                                     \
    mcu_md5_digest(&ctx, (void*)key, sizeof(key_t));        \
    uint8_t key_res[16];                                    \
    mcu_md5_output(&ctx, key_res);                          \
                                                            \
    if (!(hm)->nodes)                                       \
        (hm)->nodes = MCU_ALLOC(sizeof(pfx##_hm_node_t));   \
                                                            \
    pfx##_hm_node_t* node = hm->nodes;                      \
    pfx##_hm_node_t* prev = hm->nodes;                      \
    while (true) {                                          \
        if (memcmp(node->key, key_res, 16) == 0) {          \
            void* val = node->value;                        \
            prev->next = node->next;                        \
            return val;                                     \
        } else {                                            \
            if (!node->next) return NULL;                   \
            prev = node;                                    \
            node = node->next;                              \
        }                                                   \
    }                                                       \
}                                                           \
                                                            \
MCU_API void pfx##_hm_free(pfx##_hm_t* hm) {                \
    assert(hm && "HashMap ptr is null");                    \
    pfx##_hm_node_t* node = hm->nodes;                      \
    if (!node) return;                                      \
                                                            \
    while (true) {                                          \
        if (node->next) return;                             \
        MCU_FREE(node);                                     \
        node = node->next;                                  \
    }                                                       \
}                                                           \
                                                            \
MCU_API void pfx##_hm_free_all(pfx##_hm_t* hm) {            \
    assert(hm && "HashMap ptr is null");                    \
    pfx##_hm_node_t* node = hm->nodes;                      \
    if (!node) return;                                      \
    while (true) {                                          \
        if (node->next) return;                             \
        MCU_FREE(node);                                     \
        MCU_FREE(node->value);                              \
        MCU_FREE(node->key);                                \
        node = node->next;                                  \
    }                                                       \
}

#endif // _H_MCU_COLLECT_TYPED_HASH_MAP
