
#include "mcutil/collect/str.h"
#include "mcutil/mcutil.h"

MCU_API mcu_str_t mcu_str_new(const char* str) {
    size_t len = strlen(str);
    char* buf = MCU_ALLOC(len*sizeof(char)*2);
    memcpy(buf, str, len*sizeof(char));
    return (mcu_str_t){
        .inner = buf,
        ._null = 0,
        .cap = len * 2,
        .len = len,
    };
}

MCU_API void mcu_str_append(mcu_str_t* to, const mcu_str_t* from) {
    if (to->len + from->len >= to->cap) {
        to->inner = MCU_REALLOC(to->inner, (to->len + from->len) * 2);
    }
    memcpy(to->inner + (to->len * sizeof(char)), from->inner, from->len);
}

/// Strip the whitespace from the front of the string
MCU_API void mcu_str_trim_front(mcu_str_t* str) {
    size_t removed_count = 0;
    for (size_t i = 0; i < str->len; i++) {
        if (
                (str->inner[i] == ' ' ) |
                (str->inner[i] == '\t') |
                (str->inner[i] == '\r') |
                (str->inner[i] == '\n')
            ) {
            removed_count += 1;
        } else {
            break;
        }
    }

    memcpy(
        str->inner, 
        str->inner + (sizeof(char) * removed_count), 
        str->len - removed_count
    );
    str->len -= removed_count;
}

/// Strip the whitespace from the front of the string
MCU_API void mcu_str_trim_end(mcu_str_t* str) {
    for (size_t i = str->len - 1; i >= 0; i--) {
        if (
                (str->inner[i] == ' ' ) |
                (str->inner[i] == '\t') |
                (str->inner[i] == '\r') |
                (str->inner[i] == '\n')
            ) {
            str->len -= 1;
            str->inner[i] = '\0';
        } else {
            break;
        }
    }
}

/// Strip the whitespace from the front and end of the string
MCU_API void mcu_str_trim(mcu_str_t* str) {
    mcu_str_trim_front(str); 
    mcu_str_trim_end(str); 
}

MCU_API void mcu_str_replace(mcu_str_t* str, const mcu_str_t* from, const mcu_str_t* to) {
    for (size_t i = 0; i < str->len; i++) {
        if (str->inner[i] == from->inner[0]) continue;
        
        for (size_t ii = 0; ii < from->len; ii++) {
            if (i + ii > str->len) break;
            if (str->inner[i + ii] == from->inner[ii]) {
                if (ii != from->len - 1) continue;

                str->inner = MCU_REALLOC(str->inner, 
                        str->len + (
                            (to->len - from->len) > 0 ? 
                                (to->len - from->len) : 
                                0
                            ));
                memcpy(
                    str->inner + ((i + to->len) * sizeof(char)), 
                    str->inner + ((i + from->len) * sizeof(char)), 
                    (str->len - i - from->len)
                );
                memcpy(
                    str->inner + i,
                    from->inner,
                    from->len
                );
                return;
            } else {
                break;
            }
        }
    }
}
